# File: tello_rc.py
# Author: Michael Huelsman
# Copyright: Dr. Michael Andrew Huelsman 2023
# License: GNU GPLv3
# Created On: 29 Jun 2023
# Purpose:
#   A class for handling a multi-processed tello drone using keyboard controls.
# Notes:


from multiprocessing import Queue, Process
from queue import Empty
from time import perf_counter
from math import log1p
from pygame import display, draw, Surface, Vector2, QUIT, SRCALPHA
from pygame.font import Font, get_default_font
from pygame.image import frombuffer
from pygame import KEYDOWN, QUIT, K_l, K_t, K_DELETE, K_BACKSPACE, K_ESCAPE
from pygame import K_UP, K_DOWN, K_LEFT, K_RIGHT
from pygame.key import get_pressed, key_code
from pygame import event as pg_event
from threading import Thread
from math import sin, cos, radians

from .tello_remote import tello_remote_loop
from .tello_state import tello_state_loop
from .tello_video import tello_video_loop


class TelloRC:
    # Precond:
    #   The computer creating the TelloDrone instance is connected to the Tello's Wi-Fi.
    #
    # Postcond:
    #   Sets up a connection with the Tello Drone.
    def __init__(self):
        # Setup command process
        self.rcQ = Queue()
        self.rc_confQ = Queue()
        self.rc_process = Process(target=tello_remote_loop, args=(self.rcQ, self.rc_confQ))
        
        # Setup state process
        self.state_haltQ = Queue()
        self.state_recQ = Queue()
        self.state_process = Process(target=tello_state_loop, args=(self.state_haltQ, self.state_recQ))
        self.state_thread = Thread(target=self.__state_thread)
        
        # Setup video process
        self.video_haltQ = Queue()
        self.video_recQ = Queue()
        self.video_process = Process(target=tello_video_loop, args=(self.video_haltQ, self.video_recQ))
        self.video_thread = Thread(target=self.__video_thread)
        
        # Internal variables
        self.current_rc = [0, 0, 0, 0]
        self.last_state = None
        self.last_frame = None
        self.running = False
        self.vel_timing = 10

        self.hud_font = Font(get_default_font(), 20)
        # self.hud_thread.daemon = True
        # Create base visuals for hud
        self.hud_rad = 50
        self.hud_base = Surface((4 * self.hud_rad, 4 * self.hud_rad), SRCALPHA, 32)
        hud_center = Vector2(self.hud_base.get_width() // 2, self.hud_base.get_height() // 2)
        draw.circle(self.hud_base, (0, 0, 0, 100), hud_center, 2 * self.hud_rad)
        # Draw baselines
        draw.circle(self.hud_base, (0, 200, 0), hud_center, self.hud_rad, self.hud_rad // 10)
        # Roll baseline
        roll_baseline = Vector2(cos(0), sin(0))
        roll_start = (roll_baseline * self.hud_rad) + hud_center
        roll_end = (roll_baseline * (2 * self.hud_rad)) + hud_center
        draw.line(self.hud_base, (0, 200, 0), roll_start, roll_end, 6)
        roll_baseline - -roll_baseline
        roll_start = (roll_baseline * self.hud_rad) + hud_center
        roll_end = (roll_baseline * (2 * self.hud_rad)) + hud_center
        draw.line(self.hud_base, (0, 200, 0), roll_start, roll_end, 6)
        # Pitch indicator
        draw.line(self.hud_base, (0, 0, 220),
                  (hud_center.x - self.hud_rad + (self.hud_rad // 10), hud_center.y),
                  (hud_center.x - (self.hud_rad // 2), hud_center.y), 6)
        draw.line(self.hud_base, (0, 0, 220),
                  (hud_center.x + (self.hud_rad // 2), hud_center.y),
                  (hud_center.x + self.hud_rad - (self.hud_rad // 10), hud_center.y), 6)

    # ==========================
    #   MANAGEMENT METHODS
    # ==========================

    # Precond:
    #   None.
    #
    # Postcond:
    #   Starts the TelloRC object.
    #   Connects to the Tello.
    #   Connects to all streams and begins all processes.
    #   Returns True if all processes have been started.
    def start(self):
        self.running = True
        self.rc_process.start()
        self.state_thread.start()
        self.state_process.start()
        self.video_thread.start()
        self.video_process.start()
        return True
    
    def control(self):
        # Setup Pygame
        screen = display.set_mode((960, 720))
        display.set_caption("Tello HUD")
        # Setup
        key_holds = {'w': 0, 's': 0, 'd': 0, 'a': 0, 'q': 0, 'e': 0, 'r': 0, 'f': 0}
        poll_timer = perf_counter()
        poll_delta = 1/30
        pg_event.set_allowed([KEYDOWN])
        # Main Loop
        control_running = True
        while control_running:
            delta = perf_counter() - poll_timer
            if delta >= poll_delta:
                self.__hud_update(screen)
                poll_timer = perf_counter()
                # Check for events
                command_sent = False
                for event in pg_event.get(KEYDOWN, QUIT):
                    if event.type == QUIT:
                        control_running = False
                    if event.type == KEYDOWN:
                        if event.key == K_t:
                            command_sent = True
                            if not self.takeoff():
                                print("Problem with takeoff!")
                        elif event.key == K_l:
                            command_sent = True
                            if not self.land():
                                print("Problem with landing!")
                        elif event.key == K_ESCAPE:
                            command_sent = True
                            if not self.emergency():
                                print("Problem with emergency shutdown!")
                        elif event.key == K_BACKSPACE:
                            control_running = False
                        elif event.key == K_DELETE:
                            self.rcQ.put((0, 0, 0, 0))
                        elif event.key == K_UP:
                            command_sent = True
                            if not self.flip_forward():
                                print("Problem with forward flip!")
                        elif event.key == K_DOWN:
                            command_sent = True
                            if not self.flip_backward():
                                print("Problem with backward flip!")
                        elif event.key == K_RIGHT:
                            command_sent = True
                            if not self.flip_right():
                                print("Problem with right flip!")
                        elif event.key == K_LEFT:
                            command_sent = True
                            if not self.flip_left():
                                print("Problem with left flip!")
                if command_sent:
                    poll_timer = perf_counter()
                    continue
                # Deal with held keys
                key_state = get_pressed()
                for key in key_holds:
                    if key_state[key_code(key)]:
                        key_holds[key] += delta
                    else:
                        key_holds[key] -= delta
                    key_holds[key] = max(0, min(self.vel_timing, key_holds[key]))
                y = self.__vel_curve(key_holds['w']) - self.__vel_curve(key_holds['s'])
                x = self.__vel_curve(key_holds['d']) - self.__vel_curve(key_holds['a'])
                z = self.__vel_curve(key_holds['r']) - self.__vel_curve(key_holds['f'])
                rot = self.__vel_curve(key_holds['e']) - self.__vel_curve(key_holds['q'])
                if self.rcQ.empty():
                    self.rcQ.put((int(x), int(y), int(z), int(rot)))
    
    # Precond:
    #   None.
    #
    # Postcond:
    #   Closes down communication with the drone and writes the log to a file.
    def close(self):
        self.running = False
        if self.state_thread.is_alive():
            self.state_thread.join()
        if self.video_thread.is_alive():
            self.video_thread.join()
        if self.video_process.is_alive():
            self.video_haltQ.put("halt")
            TelloRC.__clear_q(self.video_recQ)
            self.video_process.join()
        if self.state_process.is_alive():
            self.state_haltQ.put("halt")
            TelloRC.__clear_q(self.state_recQ)
            self.state_process.join()
        if self.rc_process.is_alive():
            self.rcQ.put("halt")
            TelloRC.__clear_q(self.rc_confQ)
            self.rc_process.join()

    # ======================================
    # COMMAND METHODS
    # ======================================
    # Section Notes:
    #   All commands check to see if the drone has been connected and put into SDK mode before sending commands.

    # Precond:
    #   None.
    #
    # Postcond
    #   Sends the takeoff command.
    #   If a non-okay response is given returns False, otherwise returns true.
    def takeoff(self):
        self.rcQ.put("takeoff")
        try:
            conf = self.rc_confQ.get(block=True, timeout=5)
            if not conf[0]:
                return False
        except Empty:
            return False
        return True

    # Precond:
    #   None.
    #
    # Postcond
    #   Sends the land command.
    #   If a non-okay response is given returns False, otherwise returns true.
    def land(self):
        self.rcQ.put("land")
        try:
            conf = self.rc_confQ.get(block=True, timeout=5)
            if not conf[0]:
                return False
        except Empty:
            return False
        return True

    # Precond:
    #   None.
    #
    # Postcond:
    #   Sends the emergency command, in triplicate. Does not wait for response.
    def emergency(self):
        self.rcQ.put("emergency")
        try:
            conf = self.rc_confQ.get(block=True, timeout=5)
            if not conf[0]:
                return False
        except Empty:
            return False
        return True

    # Precond:
    #   None.
    #
    # Postcond:
    #   Adds flip l command to the command queue.
    def flip_left(self):
        self.rcQ.put("flip l")
        try:
            conf = self.rc_confQ.get(block=True, timeout=5)
            if not conf[0]:
                return False
        except Empty:
            return False
        return True

    # Precond:
    #   None.
    #
    # Postcond:
    #   Adds flip r command to the command queue.
    def flip_right(self):
        self.rcQ.put("flip r")
        try:
            conf = self.rc_confQ.get(block=True, timeout=5)
            if not conf[0]:
                return False
        except Empty:
            return False
        return True

    # Precond:
    #   None.
    #
    # Postcond:
    #   Adds flip f command to the command queue.
    def flip_forward(self):
        self.rcQ.put("flip f")
        try:
            conf = self.rc_confQ.get(block=True, timeout=5)
            if not conf[0]:
                return False
        except Empty:
            return False
        return True

    # Precond:
    #   None.
    #
    # Postcond:
    #   Adds flip b command to the command queue.
    def flip_backward(self):
        self.rcQ.put("flip b")
        try:
            conf = self.rc_confQ.get(block=True, timeout=5)
            if not conf[0]:
                return False
        except Empty:
            return False
        return True

    # ======================================
    # Info METHODS
    # ======================================

    # Precond:
    #   None.
    #
    # Postcond:
    #   Returns the last frame taken by the Tello.
    #   Returns None if the stream is off.
    def get_frame(self):
        return self.last_frame

    # Precond:
    #   None.
    #
    # Postcond:
    #   Returns the last state received from the Tello as a dictionary.
    def get_state(self):
        return self.last_state

    # ======================================
    # PRIVATE METHODS
    # ======================================

    # Precond:
    #   None.
    #
    # Postcond:
    #   Thread handling state extraction..
    def __state_thread(self):
        while self.running:
            self.last_state = self.state_recQ.get()

    # Precond:
    #   None.
    #
    # Postcond:
    #   Thread handling video extraction.
    def __video_thread(self):
        while self.running:
            self.last_frame = self.video_recQ.get()

    # Precond:
    #   q is a valid mp.Queue object.
    #
    # Postcond:
    #   Clears the q and closes it.
    def __clear_q(q: Queue):
        try:
            while not q.empty():
                q.get_nowait()
        except Empty:
            pass
        q.close()
    
    # Precond:
    #   t is a valid floating point value.
    #
    # Postcond:
    #   Returns the current velocity based on how long a key has  been pressed.
    def __vel_curve(self, t):
        return 100 * (log1p(t) / log1p(self.vel_timing))

    def __hud_update(self, screen):
        # Setup video loop basics
        horizon_rad = 50
        horizon_size = 4 * horizon_rad
        horizon_placement = ((screen.get_width() - horizon_size) // 2, (screen.get_height() - horizon_size) // 2)
        screen.fill((0, 0, 0, 255))
        if self.last_frame is not None:
            screen.blit(frombuffer(self.last_frame.tobytes(), self.last_frame.shape[1::-1], "BGR"), (0, 0))
            if self.last_state is not None:
                bat_text = self.hud_font.render(f"Battery: {self.last_state['bat']:4}", True, (0, 200, 0), (0, 0, 0))
                height_text = self.hud_font.render(f"ToF: {self.last_state['tof']:4}", True, (0, 200, 0), (0, 0, 0))
                horizon = self.__artificial_horizon(horizon_rad, int(self.last_state['pitch']), int(self.last_state['roll']))
                screen.blits([
                    (bat_text, (0, 0)),
                    (height_text, (0, bat_text.get_height())),
                    (horizon, horizon_placement)
                ])
        else:
            init_text = self.hud_font.render("Initializing...", True, (0, 200, 0), (9, 0, 0))
            x = (screen.get_width() - init_text.get_width()) // 2
            y = (screen.get_height() - init_text.get_height()) // 2
            screen.blit(init_text, (x, y))
        display.flip()

    def __artificial_horizon(self, rad: int, pitch: int, roll: int):
        result = Surface((4 * rad, 4 * rad), SRCALPHA, 32)
        result.blit(self.hud_base, (0, 0))
        center = Vector2(result.get_width() // 2, result.get_height() // 2)
        # Draw roll lines
        left_ang = radians(180 + roll)
        left_vec = Vector2(cos(left_ang), sin(left_ang))
        left_start = left_vec * rad
        left_end = left_vec * (rad * 2)
        left_start = left_start + center
        left_end = left_end + center
        draw.line(result, (0, 200, 0), left_start, left_end, 3)
        right_ang = radians(roll)
        right_vec = Vector2(cos(right_ang), sin(right_ang))
        right_start = right_vec * rad
        right_end = right_vec * rad * 2
        right_start = right_start + center
        right_end = right_end + center
        draw.line(result, (0, 200, 0), right_start, right_end, 3)
        # Draw Pitch lines
        pitch = -pitch  # Line direction adjustment
        pitch_line_deg = 10
        pitch_lines = Surface((rad, rad), SRCALPHA)
        pixels_per_ang = 2
        start_angle = pitch - ((pitch_lines.get_height() // 2) // pixels_per_ang)
        pixel_start = pitch_line_deg - (start_angle % pitch_line_deg)
        pixel_start = pixel_start * pixels_per_ang
        angles = []
        for i in range(pixel_start, pitch_lines.get_height(), pitch_line_deg * pixels_per_ang):
            ang = start_angle + (i // pixels_per_ang)
            angles.append(ang)
            if ang % pitch_line_deg == 0:
                if ang != 0:
                    draw.line(pitch_lines, (0, 200, 0), (0, i), (pitch_lines.get_width(), i), 3)
                else:
                    draw.line(pitch_lines, (200, 0, 0), (0, i), (pitch_lines.get_width(), i), 3)
        line_adj = Vector2(-(pitch_lines.get_width() // 2), -(pitch_lines.get_height() // 2))
        center_pitch = center + line_adj
        # Current Pitch indicator
        indicator_height = pitch_lines.get_height() // 2
        draw.line(pitch_lines, (0, 0, 200), (0, indicator_height), (pitch_lines.get_width(), indicator_height), 3)
        result.blit(pitch_lines, center_pitch)
        return result



# File: face_encoder.py
# Author: Michael Huelsman
# Copyright: Dr. Michael Andrew Huelsman 2023
# License: GNU GPLv3
# Created On: 29 Jun 2023
# Purpose:
#   A simple face recognition suite.
# Notes:

import pickle
from numpy import ones
from face_recognition import face_encodings, face_locations, face_distance
import cv2


class FaceEncoder:
    # Precond:
    #   encode is a string containing a file name to load a set of encodings from.
    #
    # Postcond:
    #   Creates a new FaceEncoder object loaded with encodings from a given file.
    def __init__(self, load_file=None):
        self.encodings = {}
        # Speed upgrades
        self.encode_scale = 4
        if load_file is not None:
            self.load(load_file)

    # Precond:
    #   name is a string containing the name of the person whose face is being encoded.
    #   img_file is a string containing the name of a file containing an image of a single
    #       person's face.
    #
    # Postcond:
    #   Updates the encodings based on the image file.
    #   Returns False if no face was found.
    def encode_face(self, name, img_file):
        img = self.__scale_image(cv2.imread(img_file))
        if name not in self.encodings:
            self.encodings[name] = []
        locations = face_locations(img)
        if len(locations) > 0:
            location = locations[0]
            encoding = face_encodings(img, [location], model='large')[0]
            self.encodings[name].append(encoding)
            return True
        return False
    
    # Precond:
    #   img is a valid ndarray representing an image.
    #   min_dist is a floating point number indicating the minimal face distance (default: 0.6) for recognition.
    #
    # Postcond:
    #   Returns the most likely name of the faces in the image, paired with their locations.
    #   If no face matches within the specified distance, then it is labeled Unknown.
    #   Returned tuples are (name, location)
    def detect_faces(self, img, min_dist=0.6):
        img = self.__scale_image(img)
        locations = face_locations(img)
        if len(locations) == 0:
            return []
        if len(self.encodings.keys()) == 0:
            idents = ['unknown' for i in range(len(locations))]
            locations = list(map(self.__unscale_rect, locations))
            return list(zip(idents, locations))
        encodings = face_encodings(img, locations)
        names = list(self.encodings.keys())
        distances = ones(len(names))
        idents = []
        for encoding in encodings:
            for idx, name in enumerate(names):
                distances[idx] = face_distance(self.encodings[name], encoding).mean()
            if distances.min(initial=1.0) > min_dist:
                idents.append("unknown")
            else:
                idents.append(names[distances.argmin()])
        locations = list(map(self.__unscale_rect, locations))
        return list(zip(idents, locations))

    # Precond:
    #   filename is the path to a file which contains serialized face encodings.
    #
    # Postcond:
    #   Updates the recognizers encodings using the provided file.
    # Note:
    #   No exceptions handled by this method.
    def load(self, filename):
        with open(filename, 'rb') as fin:
            loaded = pickle.load(fin)
        for person in loaded:
            if person not in self.encodings:
                self.encodings[person] = []
            self.encodings[person].extend(loaded[person])

    # Precond:
    #   filename is the path to a file to save the serialized face encodings to.
    #
    # Postcond:
    #   Saves the known face encodings to the provided file.
    # Note:
    #   No exceptions handled by this method.
    def save(self, filename):
        with open(filename, 'wb') as fout:
            pickle.dump(self.encodings, fout)
        
    # Precond:
    #   location is a rectangle on the resized image.
    #
    # Postcond:
    #   Returns a rectangle on the non-scaled image.
    def __unscale_rect(self, location):
        return tuple(map(lambda x: int(x * self.encode_scale), location))
    
    # Precond:
    #   img is a numpy array representing an image.
    #
    # Postcond:
    #   Returns a resized image by the encoder's scale factor.
    def __scale_image(self, img):
        height = 720 // self.encode_scale
        width = 960 // self.encode_scale
        return cv2.resize(img, (width, height))
    


# File: tello_remote.py
# Author: Michael Huelsman
# Copyright: Dr. Michael Andrew Huelsman 2023
# License: GNU GPLv3
# Created On: 29 Jun 2023
# Purpose:
#   A class and function for handling a multi-processed rc tello controller.
# Notes:

from socket import socket, AF_INET, SOCK_DGRAM
from threading import Thread
from time import perf_counter, sleep
from multiprocessing import Queue
from queue import Empty


def tello_remote_loop(rc_q: Queue, conf_q: Queue):
    # Create a management object
    manager = TelloRemote()
    res = manager.startup()
    conf_q.put((res, 'startup'))
    if not res:
        return
    running = True
    while running:
        # Grab from the rcQ
        if not rc_q.empty():
            current = rc_q.get()
            if type(current) == str:
                if current == "halt":
                    running = False
                elif current == "takeoff":
                    conf_q.put((manager.takeoff(), 'takeoff'))
                elif current == "land":
                    conf_q.put((manager.land(), 'land'))
                elif current == "emergency":
                    conf_q.put((manager.emergency(), 'emergency'))
                elif current == "flip f":
                    conf_q.put((manager.flip_forward(), 'flip f'))
                elif current == "flip b":
                    conf_q.put((manager.flip_backward(), 'flip b'))
                elif current == "flip r":
                    conf_q.put((manager.flip_right(), 'flip r'))
                elif current == "flip l":
                    conf_q.put((manager.flip_left(), 'flip l'))
            elif type(current) == tuple:
                manager.set_rc(*current)
    try:
        while not rc_q.empty():
            rc_q.get_nowait()
    except Empty:
        pass
    rc_q.close()
    manager.close()
    

class TelloRemote:
    def __init__(self):
        # Addresses
        self.local_addr = ('', 8889)
        self.tello_addr = ('192.168.10.1', 8889)
    
        # Setup channels
        self.send_channel = socket(AF_INET, SOCK_DGRAM)
        self.send_channel.bind(self.local_addr)
    
        self.stop = True
        self.connected = False
    
        # Setup receiving thread
        self.receive_thread = Thread(target=self.__receive)
        self.receive_thread.daemon = True
    
        # Setup logs
        self.log = []
        self.rc = [0, 0, 0, 0]
        self.MAX_TIME_OUT = 10  # measured in seconds
        self.rc_tick = 10
        self.waiting = False
        self.last_beat = perf_counter()
    
    def startup(self):
        self.stop = False
        self.receive_thread.start()
        # self.rc_beat_thread.start()
        if self.__connect(5):
            self.stream_on()
            return True
        return False
    
    # ======================================
    # COMMAND METHODS
    # ======================================

    # Precond:
    #   The value to set the forward throttle to.
    #
    # Postcond:
    #   Set forward throttle.
    def set_x(self, val):
        nv = min(max(-100, int(val)), 100)
        if nv != self.rc[1]:
            self.rc[1] = nv
            self.__send_rc()

    # Precond:
    #   The value to set the right throttle to.
    #
    # Postcond:
    #   Set right throttle.
    def set_y(self, val):
        nv = min(max(-100, int(val)), 100)
        if nv != self.rc[1]:
            self.rc[0] = nv
            self.__send_rc()

    # Precond:
    #   The value to set the vertical throttle to.
    #
    # Postcond:
    #   Set vertical throttle.
    def set_z(self, val):
        nv = min(max(-100, int(val)), 100)
        if nv != self.rc[1]:
            self.rc[2] = nv
            self.__send_rc()

    # Precond:
    #   The value to set the rotation throttle to.
    #
    # Postcond:
    #   Set rotation throttle.
    def set_rot(self, val):
        nv = min(max(-100, int(val)), 100)
        if nv != self.rc[1]:
            self.rc[3] = nv
            self.__send_rc()

    # Precond:
    #   x, y, z, and rot are numberic calues indicating the rc settings.
    #
    # Postcond:
    #   Updates the rc and sends the command if anything has changed.
    def set_rc(self, x, y, z, rot):
        nx = min(max(-100, int(x)), 100)
        ny = min(max(-100, int(y)), 100)
        nz = min(max(-100, int(z)), 100)
        nrot = min(max(-100, int(rot)), 100)
        if [nx, ny, nz, nrot] != self.rc:
            self.rc = [nx, ny, nz, nrot]
            self.__send_rc()

    # Precond:
    #   None.
    #
    # Postcond
    #   Sends the takeoff command.
    #   If a non-okay response is given returns False, otherwise returns true.
    def takeoff(self):
        if not self.connected:
            return False
        res = self.__send("takeoff")
        return res is not None and res == 'ok'

    # Precond:
    #   None.
    #
    # Postcond
    #   Sends the land command.
    #   If a non-okay response is given returns False, otherwise returns true.
    def land(self):
        if not self.connected:
            return False
        res = self.__send("land")
        return res is not None and res == 'ok'

    # Precond:
    #   None.
    #
    # Postcond:
    #   Flips the drone left.
    #   Returns False if the command could not be sent or executed for any
    #       reason.
    def flip_left(self):
        if not self.connected:
            return False
        res = self.__send("flip l")
        return res is not None and res == 'ok'

    # Precond:
    #   None.
    #
    # Postcond:
    #   Flips the drone right.
    #   Returns False if the command could not be sent or executed for any
    #       reason.
    def flip_right(self):
        if not self.connected:
            return False
        res = self.__send("flip r")
        return res is not None and res == 'ok'

    # Precond:
    #   None.
    #
    # Postcond:
    #   Flips the drone forward.
    #   Returns False if the command could not be sent or executed for any
    #       reason.
    def flip_forward(self):
        if not self.connected:
            return False
        res = self.__send("flip f")
        return res is not None and res == 'ok'

    # Precond:
    #   None.
    #
    # Postcond:
    #   Flips the drone backward.
    #   Returns False if the command could not be sent or executed for any
    #       reason.
    def flip_backward(self):
        if not self.connected:
            return False
        res = self.__send("flip b")
        return res is not None and res == 'ok'
    
    # Precond:
    #   None.
    #
    # Postcond:
    #   Sends the emergency command, in triplicate. Does not wait for response.
    def emergency(self):
        for _ in range(3):
            self.__send_nowait("emergency")
    
    # Precond:
    #   None.
    #
    # Postcond:
    #   Sends a command to the Tello turning its video stream on.
    def stream_on(self):
        res = self.__send("streamon")
        return res is not None and res == 'ok'
    
    # Precond:
    #   None.
    #
    # Postcond:
    #   Sends a command to the Tello turing its video stream off.
    def stream_off(self):
        res = self.__send("streamoff")
        return res is not None and res == 'ok'
    
    # Precond:
    #   None.
    #
    # Postcond:
    #   Closes down communication with the drone and writes the log to a file.
    def close(self):
        self.stop = True
        self.send_channel.close()
        if self.receive_thread.is_alive():
            self.receive_thread.join()
    
    # ======================================
    # PRIVATE METHODS
    # ======================================
    
    # Precond:
    #   attempts is the number of times to try and connect.
    #
    # Postcond:
    #   Checks connection to the drone by sending a message to
    #     switch the drone into SDK mode.
    #   Returns true if the connection was made.
    #   Returns false if there was a problem connecting and attempts were
    #       exceeded.
    def __connect(self, attempts=5):
        for _ in range(attempts):
            res = self.__send("command")
            if res is not None and res == 'ok':
                self.connected = True
                return True
        return False
    
    # Precond:
    #   mess is a string containing the message to send.
    #
    # Postcond:
    #   Sends the given message to the Tello.
    #   Returns the response string if the message was received.
    #   Returns None if the message failed.
    def __send(self, msg):
        self.log.append([msg, None])
        self.send_channel.sendto(msg.encode('utf-8'), self.tello_addr)
        # Response wait loop
        start = perf_counter()
        self.waiting = True
        while self.log[-1][1] is None:
            if (perf_counter() - start) > self.MAX_TIME_OUT:
                self.log[-1][1] = "TIMED OUT"
                self.waiting = False
                return None
        self.waiting = False
        return self.log[-1][1]
    
    # Precond:
    #   mess is a string containing the message to send.
    #
    # Postcond:
    #   Sends the given message to the Tello.
    #   Does not wait for a response.
    #   Used (internally) only for sending the emergency signal (which is sent
    #       in triplicate.)
    def __send_nowait(self, msg):
        self.send_channel.sendto(msg.encode('utf-8'), self.tello_addr)
        return None
    
    # Precond:
    #   None.
    #
    # Postcond:
    #   Receives messages from the Tello and logs them.
    def __receive(self):
        while not self.stop:
            try:
                response, ip = self.send_channel.recvfrom(1024)
                response = response.decode('utf-8')
                self.log[-1][1] = response.strip()
            except OSError as exc:
                if not self.stop:
                    print("Caught exception socket.error : %s" % exc)
            except UnicodeDecodeError as dec:
                if not self.stop:
                    print("Caught exception Unicode 0xcc error.")

    # Precond:
    #   None.
    #
    # Postcond:
    #   Sends RC messages to the tello based on internal throttle numbers.
    def __send_rc(self):
        cmd = 'rc ' + ' '.join(map(str, self.rc))
        self.last_beat = perf_counter()
        self.__send_nowait(cmd)


# File: __init__.py
# Author: Michael Huelsman
# Copyright: Dr. Michael Andrew Huelsman 2023
# License: GNU GPLv3
# Created On: 09 Jan 2023
# Purpose:
# Notes:
import os
os.environ['PYGAME_HIDE_SUPPORT_PROMPT'] = "hide"
import pygame as __pg
if not __pg.get_init():
    __pg.init()
from .tello_drone import TelloDrone
from .tello_hud import TelloHud
from .tello_rc import TelloRC
from .face_encoder import FaceEncoder
from .tello_face_hud import TelloFaceHud



# File: tello_hud.py
# Author: Michael Huelsman
# Copyright: Dr. Michael Andrew Huelsman 2023
# License: GNU GPLv3
# Created On: 27 Jun 2023
# Purpose:
#   A simple hud program that can be run on a separate thread.
# Notes:

from .tello_drone import TelloDrone
from .tello_rc import TelloRC
from time import perf_counter, sleep
from pygame import display, draw, event, Surface, Vector2, QUIT, SRCALPHA, KEYDOWN, K_p
from pygame.font import Font, get_default_font
from pygame.image import frombuffer
from math import sin, cos, radians
from threading import Thread
import uuid
from cv2 import imwrite


class TelloHud:
    def __init__(self, drone: TelloDrone | TelloRC):
        self.drone = drone
        self.running = False
        self.hud_thread = Thread(target=self.__hud_stream)
        self.hud_thread.daemon = True
        self.hud_font = Font(get_default_font(), 20)
        # self.hud_thread.daemon = True
        # Create base visuals for hud
        self.hud_rad = 50
        self.hud_base = Surface((4 * self.hud_rad, 4 * self.hud_rad), SRCALPHA, 32)
        hud_center = Vector2(self.hud_base.get_width() // 2, self.hud_base.get_height() // 2)
        draw.circle(self.hud_base, (0, 0, 0, 100), hud_center, 2 * self.hud_rad)
        # Draw baselines
        draw.circle(self.hud_base, (0, 200, 0), hud_center, self.hud_rad, self.hud_rad // 10)
        # Roll baseline
        roll_baseline = Vector2(cos(0), sin(0))
        roll_start = (roll_baseline * self.hud_rad) + hud_center
        roll_end = (roll_baseline * (2 * self.hud_rad)) + hud_center
        draw.line(self.hud_base, (0, 200, 0), roll_start, roll_end, 6)
        roll_baseline - -roll_baseline
        roll_start = (roll_baseline * self.hud_rad) + hud_center
        roll_end = (roll_baseline * (2 * self.hud_rad)) + hud_center
        draw.line(self.hud_base, (0, 200, 0), roll_start, roll_end, 6)
        # Pitch indicator
        draw.line(self.hud_base, (0, 0, 220),
                  (hud_center.x - self.hud_rad + (self.hud_rad // 10), hud_center.y),
                  (hud_center.x - (self.hud_rad // 2), hud_center.y), 6)
        draw.line(self.hud_base, (0, 0, 220),
                  (hud_center.x + (self.hud_rad // 2), hud_center.y),
                  (hud_center.x + self.hud_rad - (self.hud_rad // 10), hud_center.y), 6)
    
    def activate_hud(self):
        if self.running:
            return
        self.running = True
        self.hud_thread.start()
        while self.drone.get_frame() is None:
            sleep(1)
    
    def deactivate_hud(self):
        if not self.running:
            if self.hud_thread.is_alive():
                self.hud_thread.join()
            return
        self.running = False
        self.hud_thread.join()
    
    def is_active(self):
        return self.running
    
    def __hud_stream(self):
        # Setup Pygame
        screen = display.set_mode((960, 720))
        display.set_caption("Tello HUD")
        # Setup video loop basics
        frame_timer = perf_counter()
        frame_delta = 1 / 30
        event.set_allowed([QUIT, KEYDOWN])
        horizon_rad = 50
        horizon_size = 4 * horizon_rad
        horizon_placement = ((screen.get_width() - horizon_size) // 2, (screen.get_height() - horizon_size) // 2)
        while self.running:
            delta = perf_counter() - frame_timer
            if delta >= frame_delta:
                frame_timer = perf_counter()
                screen.fill((0, 0, 0, 255))
                frame = self.drone.get_frame()
                if frame is not None:
                    screen.blit(frombuffer(frame.tobytes(), frame.shape[1::-1], "BGR"), (0, 0))
                    state = self.drone.get_state()
                    if state is not None:
                        fps_text = self.hud_font.render(f"FPS: {int(1 / delta):4}", True, (0, 200, 0), (0, 0, 0))
                        bat_text = self.hud_font.render(f"Battery: {state['bat']:4}", True, (0, 200, 0), (0, 0, 0))
                        height_text = self.hud_font.render(f"ToF: {state['tof']:4}", True, (0, 200, 0), (0, 0, 0))
                        horizon = self.__artificial_horizon(horizon_rad, int(state['pitch']), int(state['roll']))
                        screen.blits([
                            (fps_text, (0, 0)),
                            (bat_text, (0, fps_text.get_height())),
                            (height_text, (bat_text.get_width(), fps_text.get_height())),
                            (horizon, horizon_placement)
                        ])
                else:
                    init_text = self.hud_font.render("Initializing...", True, (0, 200, 0), (9, 0, 0))
                    x = (screen.get_width() - init_text.get_width()) // 2
                    y = (screen.get_height() - init_text.get_height()) // 2
                    screen.blit(init_text, (x, y))
                display.flip()
                for _ in event.get(QUIT):
                    self.running = False
                for evt in event.get(KEYDOWN):
                    if evt.key == K_p:
                        filename = str(uuid.uuid4()) + '.jpg'
                        imwrite(filename, frame)
                        
    
    def __artificial_horizon(self, rad: int, pitch: int, roll: int):
        result = Surface((4 * rad, 4 * rad), SRCALPHA, 32)
        result.blit(self.hud_base, (0, 0))
        center = Vector2(result.get_width() // 2, result.get_height() // 2)
        # Draw roll lines
        left_ang = radians(180 + roll)
        left_vec = Vector2(cos(left_ang), sin(left_ang))
        left_start = left_vec * rad
        left_end = left_vec * (rad * 2)
        left_start = left_start + center
        left_end = left_end + center
        draw.line(result, (0, 200, 0), left_start, left_end, 3)
        right_ang = radians(roll)
        right_vec = Vector2(cos(right_ang), sin(right_ang))
        right_start = right_vec * rad
        right_end = right_vec * rad * 2
        right_start = right_start + center
        right_end = right_end + center
        draw.line(result, (0, 200, 0), right_start, right_end, 3)
        # Draw Pitch lines
        pitch = -pitch  # Line direction adjustment
        pitch_line_deg = 10
        pitch_lines = Surface((rad, rad), SRCALPHA)
        pixels_per_ang = 2
        start_angle = pitch - ((pitch_lines.get_height() // 2) // pixels_per_ang)
        pixel_start = pitch_line_deg - (start_angle % pitch_line_deg)
        pixel_start = pixel_start * pixels_per_ang
        angles = []
        for i in range(pixel_start, pitch_lines.get_height(), pitch_line_deg * pixels_per_ang):
            ang = start_angle + (i // pixels_per_ang)
            angles.append(ang)
            if ang % pitch_line_deg == 0:
                if ang != 0:
                    draw.line(pitch_lines, (0, 200, 0), (0, i), (pitch_lines.get_width(), i), 3)
                else:
                    draw.line(pitch_lines, (200, 0, 0), (0, i), (pitch_lines.get_width(), i), 3)
        line_adj = Vector2(-(pitch_lines.get_width() // 2), -(pitch_lines.get_height() // 2))
        center_pitch = center + line_adj
        # Current Pitch indicator
        indicator_height = pitch_lines.get_height() // 2
        draw.line(pitch_lines, (0, 0, 200), (0, indicator_height), (pitch_lines.get_width(), indicator_height), 3)
        result.blit(pitch_lines, center_pitch)
        return result



# File: tello_face_hud.py
# Author: Michael Huelsman
# Copyright: Dr. Michael Andrew Huelsman 2023
# License: GNU GPLv3
# Created On: 29 Jun 2023
# Purpose:
# Notes:

from .tello_drone import TelloDrone
from .tello_rc import TelloRC
from .face_encoder import FaceEncoder
from time import perf_counter
from pygame import display, draw, event, Rect, DOUBLEBUF, OPENGLBLIT, QUIT
from pygame.font import Font, get_default_font
from pygame.image import frombuffer
from threading import Thread


class TelloFaceHud:
    def __init__(self, drone: TelloDrone | TelloRC, faces: FaceEncoder):
        self.drone = drone
        self.encoder = faces
        self.running = False
        self.hud_thread = Thread(target=self.__hud_stream)
        self.hud_thread.daemon = True
        
        #pg reqs
        self.font = Font(get_default_font(), 16)
    
    def activate_hud(self):
        if self.running:
            return
        self.running = True
        self.hud_thread.start()
    
    def deactivate_hud(self):
        if not self.running:
            if self.hud_thread.is_alive():
                self.hud_thread.join()
            return
        self.running = False
        self.hud_thread.join()
    
    def is_active(self):
        return self.running
    
    def __hud_stream(self):
        # Setup Pygame
        screen = display.set_mode((960, 720))
        display.set_caption("Tello HUD")
        # Setup video loop basics
        frame_timer = perf_counter()
        frame_delta = 1 / 30
        while self.running:
            if (perf_counter() - frame_timer) > frame_delta:
                frame_timer = perf_counter()
                drone_frame = self.drone.get_frame()
                if drone_frame is not None:
                    frame = frombuffer(drone_frame.tobytes(), drone_frame.shape[1::-1], "BGR")
                    screen.blit(frame, (0, 0))
                    # Detect Faces
                    faces = self.encoder.detect_faces(drone_frame)
                    for name, location in faces:
                        rect = TelloFaceHud.__convert_rect(location)
                        draw.rect(screen, (0, 200, 0), rect, 5)
                        name_text = self.font.render(name, True, (0, 200, 0))
                        screen.blit(name_text, (rect.x, rect.y-name_text.get_height()-1))
                display.flip()
                for _ in event.get(QUIT):
                    self.running = False
        display.quit()
    
    @staticmethod
    def __convert_rect(rect):
        left = rect[3]
        top = rect[0]
        width = rect[1]-rect[3]
        height = rect[2]-rect[0]
        return Rect(left, top, width, height)
    